/* packet-lpp.h
 * Routines for 3GPP LTE Positioning Protocol (LPP) packet dissection
 * Copyright 2011-2019 Pascal Quantin <pascal@wireshark.org>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Ref 3GPP TS 37.355 version 15.0.0 Release 15
 * http://www.3gpp.org
 */

#ifndef PACKET_LPP_H
#define PACKET_LPP_H

#include "packet-lpp-exp.h"

#endif  /* PACKET_LPP_H */
