/* packet-rwall.c
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#define NEW_PROTO_TREE_API

#include "config.h"

#include "packet-rpc.h"

void proto_register_rwall(void);
void proto_reg_handoff_rwall(void);

static header_field_info *hfi_rwall = NULL;

/* there is no procedure 1 */
#define RWALL_WALL 2

#define RWALL_PROGRAM 100008

#define RWALL_HFI_INIT HFI_INIT(proto_rwall)

static const value_string rwall1_proc_vals[] = {
	{ RWALL_WALL,	"RWALL" },
	{ 0,	NULL }
};

static header_field_info hfi_rwall_procedure_v1 RWALL_HFI_INIT = {
	"V1 Procedure", "rwall.procedure_v1", FT_UINT32, BASE_DEC,
	VALS(rwall1_proc_vals), 0, NULL, HFILL };

static header_field_info hfi_rwall_message RWALL_HFI_INIT = {
	"Message", "rwall.message", FT_STRING, BASE_NONE,
	NULL, 0, NULL, HFILL };

static gint ett_rwall = -1;

static int
dissect_rwall_call(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *tree, void* data _U_)
{
	return dissect_rpc_string(tvb, tree, hfi_rwall_message.id, 0, NULL);
}

static const vsff rwall1_proc[] = {
	{ RWALL_WALL,	"RWALL", dissect_rwall_call,	dissect_rpc_void },
	{ 0,	NULL,	NULL,	NULL }
};

static const rpc_prog_vers_info rwall_vers_info[] = {
	{ 1, rwall1_proc, &hfi_rwall_procedure_v1.id },
};

void
proto_register_rwall(void)
{
#ifndef HAVE_HFI_SECTION_INIT
	static header_field_info *hfi[] = {
		&hfi_rwall_procedure_v1,
		&hfi_rwall_message,
	};
#endif

	static gint *ett[] = {
		&ett_rwall,
	};

	int proto_rwall;

	proto_rwall = proto_register_protocol("Remote Wall protocol", "RWALL", "rwall");
	hfi_rwall = proto_registrar_get_nth(proto_rwall);

	proto_register_fields(proto_rwall, hfi, array_length(hfi));
	proto_register_subtree_array(ett, array_length(ett));
}

void
proto_reg_handoff_rwall(void)
{
	/* Register the protocol as RPC */
	rpc_init_prog(hfi_rwall->id, RWALL_PROGRAM, ett_rwall,
	    G_N_ELEMENTS(rwall_vers_info), rwall_vers_info);
}

/*
 * Editor modelines  -  https://www.wireshark.org/tools/modelines.html
 *
 * Local variables:
 * c-basic-offset: 8
 * tab-width: 8
 * indent-tabs-mode: t
 * End:
 *
 * vi: set shiftwidth=8 tabstop=8 noexpandtab:
 * :indentSize=8:tabSize=8:noTabs=false:
 */
