/******************************************************************************
** Copyright (C) 2006-2015 ascolab GmbH. All Rights Reserved.
** Web: http://www.ascolab.com
**
** SPDX-License-Identifier: GPL-2.0-or-later
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Project: OpcUa Wireshark Plugin
**
** Description: OpcUa Extension Object IDs
**
** This file was autogenerated on 13.10.2015.
** DON'T MODIFY THIS FILE!
**
******************************************************************************/

#define OpcUaId_TrustListDataType_Encoding_DefaultBinary 12680
#define OpcUaId_Node_Encoding_DefaultBinary 260
#define OpcUaId_InstanceNode_Encoding_DefaultBinary 11889
#define OpcUaId_TypeNode_Encoding_DefaultBinary 11890
#define OpcUaId_ObjectNode_Encoding_DefaultBinary 263
#define OpcUaId_ObjectTypeNode_Encoding_DefaultBinary 266
#define OpcUaId_VariableNode_Encoding_DefaultBinary 269
#define OpcUaId_VariableTypeNode_Encoding_DefaultBinary 272
#define OpcUaId_ReferenceTypeNode_Encoding_DefaultBinary 275
#define OpcUaId_MethodNode_Encoding_DefaultBinary 278
#define OpcUaId_ViewNode_Encoding_DefaultBinary 281
#define OpcUaId_DataTypeNode_Encoding_DefaultBinary 284
#define OpcUaId_ReferenceNode_Encoding_DefaultBinary 287
#define OpcUaId_Argument_Encoding_DefaultBinary 298
#define OpcUaId_EnumValueType_Encoding_DefaultBinary 8251
#define OpcUaId_OptionSet_Encoding_DefaultBinary 12765
#define OpcUaId_TimeZoneDataType_Encoding_DefaultBinary 8917
#define OpcUaId_ApplicationDescription_Encoding_DefaultBinary 310
#define OpcUaId_RequestHeader_Encoding_DefaultBinary 391
#define OpcUaId_ResponseHeader_Encoding_DefaultBinary 394
#define OpcUaId_ServerOnNetwork_Encoding_DefaultBinary 12207
#define OpcUaId_UserTokenPolicy_Encoding_DefaultBinary 306
#define OpcUaId_EndpointDescription_Encoding_DefaultBinary 314
#define OpcUaId_RegisteredServer_Encoding_DefaultBinary 434
#define OpcUaId_MdnsDiscoveryConfiguration_Encoding_DefaultBinary 12901
#define OpcUaId_ChannelSecurityToken_Encoding_DefaultBinary 443
#define OpcUaId_SignedSoftwareCertificate_Encoding_DefaultBinary 346
#define OpcUaId_SignatureData_Encoding_DefaultBinary 458
#define OpcUaId_UserIdentityToken_Encoding_DefaultBinary 318
#define OpcUaId_AnonymousIdentityToken_Encoding_DefaultBinary 321
#define OpcUaId_UserNameIdentityToken_Encoding_DefaultBinary 324
#define OpcUaId_X509IdentityToken_Encoding_DefaultBinary 327
#define OpcUaId_KerberosIdentityToken_Encoding_DefaultBinary 12509
#define OpcUaId_IssuedIdentityToken_Encoding_DefaultBinary 940
#define OpcUaId_NodeAttributes_Encoding_DefaultBinary 351
#define OpcUaId_ObjectAttributes_Encoding_DefaultBinary 354
#define OpcUaId_VariableAttributes_Encoding_DefaultBinary 357
#define OpcUaId_MethodAttributes_Encoding_DefaultBinary 360
#define OpcUaId_ObjectTypeAttributes_Encoding_DefaultBinary 363
#define OpcUaId_VariableTypeAttributes_Encoding_DefaultBinary 366
#define OpcUaId_ReferenceTypeAttributes_Encoding_DefaultBinary 369
#define OpcUaId_DataTypeAttributes_Encoding_DefaultBinary 372
#define OpcUaId_ViewAttributes_Encoding_DefaultBinary 375
#define OpcUaId_AddNodesItem_Encoding_DefaultBinary 378
#define OpcUaId_AddNodesResult_Encoding_DefaultBinary 485
#define OpcUaId_AddReferencesItem_Encoding_DefaultBinary 381
#define OpcUaId_DeleteNodesItem_Encoding_DefaultBinary 384
#define OpcUaId_DeleteReferencesItem_Encoding_DefaultBinary 387
#define OpcUaId_ViewDescription_Encoding_DefaultBinary 513
#define OpcUaId_BrowseDescription_Encoding_DefaultBinary 516
#define OpcUaId_ReferenceDescription_Encoding_DefaultBinary 520
#define OpcUaId_BrowseResult_Encoding_DefaultBinary 524
#define OpcUaId_RelativePathElement_Encoding_DefaultBinary 539
#define OpcUaId_RelativePath_Encoding_DefaultBinary 542
#define OpcUaId_BrowsePath_Encoding_DefaultBinary 545
#define OpcUaId_BrowsePathTarget_Encoding_DefaultBinary 548
#define OpcUaId_BrowsePathResult_Encoding_DefaultBinary 551
#define OpcUaId_EndpointConfiguration_Encoding_DefaultBinary 333
#define OpcUaId_SupportedProfile_Encoding_DefaultBinary 337
#define OpcUaId_SoftwareCertificate_Encoding_DefaultBinary 343
#define OpcUaId_QueryDataDescription_Encoding_DefaultBinary 572
#define OpcUaId_NodeTypeDescription_Encoding_DefaultBinary 575
#define OpcUaId_QueryDataSet_Encoding_DefaultBinary 579
#define OpcUaId_NodeReference_Encoding_DefaultBinary 582
#define OpcUaId_ContentFilterElement_Encoding_DefaultBinary 585
#define OpcUaId_ContentFilter_Encoding_DefaultBinary 588
#define OpcUaId_ElementOperand_Encoding_DefaultBinary 594
#define OpcUaId_LiteralOperand_Encoding_DefaultBinary 597
#define OpcUaId_AttributeOperand_Encoding_DefaultBinary 600
#define OpcUaId_SimpleAttributeOperand_Encoding_DefaultBinary 603
#define OpcUaId_ContentFilterElementResult_Encoding_DefaultBinary 606
#define OpcUaId_ContentFilterResult_Encoding_DefaultBinary 609
#define OpcUaId_ParsingResult_Encoding_DefaultBinary 612
#define OpcUaId_ReadValueId_Encoding_DefaultBinary 628
#define OpcUaId_HistoryReadValueId_Encoding_DefaultBinary 637
#define OpcUaId_HistoryReadResult_Encoding_DefaultBinary 640
#define OpcUaId_ReadEventDetails_Encoding_DefaultBinary 646
#define OpcUaId_ReadRawModifiedDetails_Encoding_DefaultBinary 649
#define OpcUaId_ReadProcessedDetails_Encoding_DefaultBinary 652
#define OpcUaId_ReadAtTimeDetails_Encoding_DefaultBinary 655
#define OpcUaId_HistoryData_Encoding_DefaultBinary 658
#define OpcUaId_ModificationInfo_Encoding_DefaultBinary 11226
#define OpcUaId_HistoryModifiedData_Encoding_DefaultBinary 11227
#define OpcUaId_HistoryEvent_Encoding_DefaultBinary 661
#define OpcUaId_WriteValue_Encoding_DefaultBinary 670
#define OpcUaId_HistoryUpdateDetails_Encoding_DefaultBinary 679
#define OpcUaId_UpdateDataDetails_Encoding_DefaultBinary 682
#define OpcUaId_UpdateStructureDataDetails_Encoding_DefaultBinary 11300
#define OpcUaId_UpdateEventDetails_Encoding_DefaultBinary 685
#define OpcUaId_DeleteRawModifiedDetails_Encoding_DefaultBinary 688
#define OpcUaId_DeleteAtTimeDetails_Encoding_DefaultBinary 691
#define OpcUaId_DeleteEventDetails_Encoding_DefaultBinary 694
#define OpcUaId_HistoryUpdateResult_Encoding_DefaultBinary 697
#define OpcUaId_CallMethodRequest_Encoding_DefaultBinary 706
#define OpcUaId_CallMethodResult_Encoding_DefaultBinary 709
#define OpcUaId_DataChangeFilter_Encoding_DefaultBinary 724
#define OpcUaId_EventFilter_Encoding_DefaultBinary 727
#define OpcUaId_AggregateConfiguration_Encoding_DefaultBinary 950
#define OpcUaId_AggregateFilter_Encoding_DefaultBinary 730
#define OpcUaId_EventFilterResult_Encoding_DefaultBinary 736
#define OpcUaId_AggregateFilterResult_Encoding_DefaultBinary 739
#define OpcUaId_MonitoringParameters_Encoding_DefaultBinary 742
#define OpcUaId_MonitoredItemCreateRequest_Encoding_DefaultBinary 745
#define OpcUaId_MonitoredItemCreateResult_Encoding_DefaultBinary 748
#define OpcUaId_MonitoredItemModifyRequest_Encoding_DefaultBinary 757
#define OpcUaId_MonitoredItemModifyResult_Encoding_DefaultBinary 760
#define OpcUaId_NotificationMessage_Encoding_DefaultBinary 805
#define OpcUaId_DataChangeNotification_Encoding_DefaultBinary 811
#define OpcUaId_MonitoredItemNotification_Encoding_DefaultBinary 808
#define OpcUaId_EventNotificationList_Encoding_DefaultBinary 916
#define OpcUaId_EventFieldList_Encoding_DefaultBinary 919
#define OpcUaId_HistoryEventFieldList_Encoding_DefaultBinary 922
#define OpcUaId_StatusChangeNotification_Encoding_DefaultBinary 820
#define OpcUaId_SubscriptionAcknowledgement_Encoding_DefaultBinary 823
#define OpcUaId_TransferResult_Encoding_DefaultBinary 838
#define OpcUaId_ScalarTestType_Encoding_DefaultBinary 401
#define OpcUaId_ArrayTestType_Encoding_DefaultBinary 404
#define OpcUaId_CompositeTestType_Encoding_DefaultBinary 407
#define OpcUaId_BuildInfo_Encoding_DefaultBinary 340
#define OpcUaId_RedundantServerDataType_Encoding_DefaultBinary 855
#define OpcUaId_EndpointUrlListDataType_Encoding_DefaultBinary 11957
#define OpcUaId_NetworkGroupDataType_Encoding_DefaultBinary 11958
#define OpcUaId_SamplingIntervalDiagnosticsDataType_Encoding_DefaultBinary 858
#define OpcUaId_ServerDiagnosticsSummaryDataType_Encoding_DefaultBinary 861
#define OpcUaId_ServerStatusDataType_Encoding_DefaultBinary 864
#define OpcUaId_SessionDiagnosticsDataType_Encoding_DefaultBinary 867
#define OpcUaId_SessionSecurityDiagnosticsDataType_Encoding_DefaultBinary 870
#define OpcUaId_ServiceCounterDataType_Encoding_DefaultBinary 873
#define OpcUaId_StatusResult_Encoding_DefaultBinary 301
#define OpcUaId_SubscriptionDiagnosticsDataType_Encoding_DefaultBinary 876
#define OpcUaId_ModelChangeStructureDataType_Encoding_DefaultBinary 879
#define OpcUaId_SemanticChangeStructureDataType_Encoding_DefaultBinary 899
#define OpcUaId_Range_Encoding_DefaultBinary 886
#define OpcUaId_EUInformation_Encoding_DefaultBinary 889
#define OpcUaId_ComplexNumberType_Encoding_DefaultBinary 12181
#define OpcUaId_DoubleComplexNumberType_Encoding_DefaultBinary 12182
#define OpcUaId_AxisInformation_Encoding_DefaultBinary 12089
#define OpcUaId_XVType_Encoding_DefaultBinary 12090
#define OpcUaId_ProgramDiagnosticDataType_Encoding_DefaultBinary 896
#define OpcUaId_Annotation_Encoding_DefaultBinary 893

