/******************************************************************************
** Copyright (C) 2006-2015 ascolab GmbH. All Rights Reserved.
** Web: http://www.ascolab.com
**
** SPDX-License-Identifier: GPL-2.0-or-later
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Project: OpcUa Wireshark Plugin
**
** Description: OpcUa Service IDs
**
** This file was autogenerated on 13.10.2015.
** DON'T MODIFY THIS FILE!
** XXX - well, except that you may have to.  See the README.
**
******************************************************************************/

#define OpcUaId_ServiceFault_Encoding_DefaultBinary 397
#define OpcUaId_FindServersRequest_Encoding_DefaultBinary 422
#define OpcUaId_FindServersResponse_Encoding_DefaultBinary 425
#define OpcUaId_FindServersOnNetworkRequest_Encoding_DefaultBinary 12208
#define OpcUaId_FindServersOnNetworkResponse_Encoding_DefaultBinary 12209
#define OpcUaId_GetEndpointsRequest_Encoding_DefaultBinary 428
#define OpcUaId_GetEndpointsResponse_Encoding_DefaultBinary 431
#define OpcUaId_RegisterServerRequest_Encoding_DefaultBinary 437
#define OpcUaId_RegisterServerResponse_Encoding_DefaultBinary 440
#define OpcUaId_RegisterServer2Request_Encoding_DefaultBinary 12211
#define OpcUaId_RegisterServer2Response_Encoding_DefaultBinary 12212
#define OpcUaId_OpenSecureChannelRequest_Encoding_DefaultBinary 446
#define OpcUaId_OpenSecureChannelResponse_Encoding_DefaultBinary 449
#define OpcUaId_CloseSecureChannelRequest_Encoding_DefaultBinary 452
#define OpcUaId_CloseSecureChannelResponse_Encoding_DefaultBinary 455
#define OpcUaId_CreateSessionRequest_Encoding_DefaultBinary 461
#define OpcUaId_CreateSessionResponse_Encoding_DefaultBinary 464
#define OpcUaId_ActivateSessionRequest_Encoding_DefaultBinary 467
#define OpcUaId_ActivateSessionResponse_Encoding_DefaultBinary 470
#define OpcUaId_CloseSessionRequest_Encoding_DefaultBinary 473
#define OpcUaId_CloseSessionResponse_Encoding_DefaultBinary 476
#define OpcUaId_CancelRequest_Encoding_DefaultBinary 479
#define OpcUaId_CancelResponse_Encoding_DefaultBinary 482
#define OpcUaId_AddNodesRequest_Encoding_DefaultBinary 488
#define OpcUaId_AddNodesResponse_Encoding_DefaultBinary 491
#define OpcUaId_AddReferencesRequest_Encoding_DefaultBinary 494
#define OpcUaId_AddReferencesResponse_Encoding_DefaultBinary 497
#define OpcUaId_DeleteNodesRequest_Encoding_DefaultBinary 500
#define OpcUaId_DeleteNodesResponse_Encoding_DefaultBinary 503
#define OpcUaId_DeleteReferencesRequest_Encoding_DefaultBinary 506
#define OpcUaId_DeleteReferencesResponse_Encoding_DefaultBinary 509
#define OpcUaId_BrowseRequest_Encoding_DefaultBinary 527
#define OpcUaId_BrowseResponse_Encoding_DefaultBinary 530
#define OpcUaId_BrowseNextRequest_Encoding_DefaultBinary 533
#define OpcUaId_BrowseNextResponse_Encoding_DefaultBinary 536
#define OpcUaId_TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultBinary 554
#define OpcUaId_TranslateBrowsePathsToNodeIdsResponse_Encoding_DefaultBinary 557
#define OpcUaId_RegisterNodesRequest_Encoding_DefaultBinary 560
#define OpcUaId_RegisterNodesResponse_Encoding_DefaultBinary 563
#define OpcUaId_UnregisterNodesRequest_Encoding_DefaultBinary 566
#define OpcUaId_UnregisterNodesResponse_Encoding_DefaultBinary 569
#define OpcUaId_QueryFirstRequest_Encoding_DefaultBinary 615
#define OpcUaId_QueryFirstResponse_Encoding_DefaultBinary 618
#define OpcUaId_QueryNextRequest_Encoding_DefaultBinary 621
#define OpcUaId_QueryNextResponse_Encoding_DefaultBinary 624
#define OpcUaId_ReadRequest_Encoding_DefaultBinary 631
#define OpcUaId_ReadResponse_Encoding_DefaultBinary 634
#define OpcUaId_HistoryReadRequest_Encoding_DefaultBinary 664
#define OpcUaId_HistoryReadResponse_Encoding_DefaultBinary 667
#define OpcUaId_WriteRequest_Encoding_DefaultBinary 673
#define OpcUaId_WriteResponse_Encoding_DefaultBinary 676
#define OpcUaId_HistoryUpdateRequest_Encoding_DefaultBinary 700
#define OpcUaId_HistoryUpdateResponse_Encoding_DefaultBinary 703
#define OpcUaId_CallRequest_Encoding_DefaultBinary 712
#define OpcUaId_CallResponse_Encoding_DefaultBinary 715
#define OpcUaId_CreateMonitoredItemsRequest_Encoding_DefaultBinary 751
#define OpcUaId_CreateMonitoredItemsResponse_Encoding_DefaultBinary 754
#define OpcUaId_ModifyMonitoredItemsRequest_Encoding_DefaultBinary 763
#define OpcUaId_ModifyMonitoredItemsResponse_Encoding_DefaultBinary 766
#define OpcUaId_SetMonitoringModeRequest_Encoding_DefaultBinary 769
#define OpcUaId_SetMonitoringModeResponse_Encoding_DefaultBinary 772
#define OpcUaId_SetTriggeringRequest_Encoding_DefaultBinary 775
#define OpcUaId_SetTriggeringResponse_Encoding_DefaultBinary 778
#define OpcUaId_DeleteMonitoredItemsRequest_Encoding_DefaultBinary 781
#define OpcUaId_DeleteMonitoredItemsResponse_Encoding_DefaultBinary 784
#define OpcUaId_CreateSubscriptionRequest_Encoding_DefaultBinary 787
#define OpcUaId_CreateSubscriptionResponse_Encoding_DefaultBinary 790
#define OpcUaId_ModifySubscriptionRequest_Encoding_DefaultBinary 793
#define OpcUaId_ModifySubscriptionResponse_Encoding_DefaultBinary 796
#define OpcUaId_SetPublishingModeRequest_Encoding_DefaultBinary 799
#define OpcUaId_SetPublishingModeResponse_Encoding_DefaultBinary 802
#define OpcUaId_PublishRequest_Encoding_DefaultBinary 826
#define OpcUaId_PublishResponse_Encoding_DefaultBinary 829
#define OpcUaId_RepublishRequest_Encoding_DefaultBinary 832
#define OpcUaId_RepublishResponse_Encoding_DefaultBinary 835
#define OpcUaId_TransferSubscriptionsRequest_Encoding_DefaultBinary 841
#define OpcUaId_TransferSubscriptionsResponse_Encoding_DefaultBinary 844
#define OpcUaId_DeleteSubscriptionsRequest_Encoding_DefaultBinary 847
#define OpcUaId_DeleteSubscriptionsResponse_Encoding_DefaultBinary 850
#define OpcUaId_TestStackRequest_Encoding_DefaultBinary 410
#define OpcUaId_TestStackResponse_Encoding_DefaultBinary 413
#define OpcUaId_TestStackExRequest_Encoding_DefaultBinary 416
#define OpcUaId_TestStackExResponse_Encoding_DefaultBinary 419

#define OpcUaId_ServiceFault_Encoding_DefaultXml 396
#define OpcUaId_FindServersRequest_Encoding_DefaultXml 421
#define OpcUaId_FindServersResponse_Encoding_DefaultXml 424
#define OpcUaId_FindServersOnNetworkRequest_Encoding_DefaultXml 12196
#define OpcUaId_FindServersOnNetworkResponse_Encoding_DefaultXml 12197
#define OpcUaId_GetEndpointsRequest_Encoding_DefaultXml 427
#define OpcUaId_GetEndpointsResponse_Encoding_DefaultXml 430
#define OpcUaId_RegisterServerRequest_Encoding_DefaultXml 436
#define OpcUaId_RegisterServerResponse_Encoding_DefaultXml 439
#define OpcUaId_RegisterServer2Request_Encoding_DefaultXml 12199
#define OpcUaId_RegisterServer2Response_Encoding_DefaultXml 12200
#define OpcUaId_OpenSecureChannelRequest_Encoding_DefaultXml 445
#define OpcUaId_OpenSecureChannelResponse_Encoding_DefaultXml 448
#define OpcUaId_CloseSecureChannelRequest_Encoding_DefaultXml 451
#define OpcUaId_CloseSecureChannelResponse_Encoding_DefaultXml 454
#define OpcUaId_CreateSessionRequest_Encoding_DefaultXml 460
#define OpcUaId_CreateSessionResponse_Encoding_DefaultXml 463
#define OpcUaId_ActivateSessionRequest_Encoding_DefaultXml 466
#define OpcUaId_ActivateSessionResponse_Encoding_DefaultXml 469
#define OpcUaId_CloseSessionRequest_Encoding_DefaultXml 472
#define OpcUaId_CloseSessionResponse_Encoding_DefaultXml 475
#define OpcUaId_CancelRequest_Encoding_DefaultXml 478
#define OpcUaId_CancelResponse_Encoding_DefaultXml 481
#define OpcUaId_AddNodesRequest_Encoding_DefaultXml 487
#define OpcUaId_AddNodesResponse_Encoding_DefaultXml 490
#define OpcUaId_AddReferencesRequest_Encoding_DefaultXml 493
#define OpcUaId_AddReferencesResponse_Encoding_DefaultXml 496
#define OpcUaId_DeleteNodesRequest_Encoding_DefaultXml 499
#define OpcUaId_DeleteNodesResponse_Encoding_DefaultXml 502
#define OpcUaId_DeleteReferencesRequest_Encoding_DefaultXml 505
#define OpcUaId_DeleteReferencesResponse_Encoding_DefaultXml 508
#define OpcUaId_BrowseRequest_Encoding_DefaultXml 526
#define OpcUaId_BrowseResponse_Encoding_DefaultXml 529
#define OpcUaId_BrowseNextRequest_Encoding_DefaultXml 532
#define OpcUaId_BrowseNextResponse_Encoding_DefaultXml 535
#define OpcUaId_TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultXml 553
#define OpcUaId_TranslateBrowsePathsToNodeIdsResponse_Encoding_DefaultXml 556
#define OpcUaId_RegisterNodesRequest_Encoding_DefaultXml 559
#define OpcUaId_RegisterNodesResponse_Encoding_DefaultXml 562
#define OpcUaId_UnregisterNodesRequest_Encoding_DefaultXml 565
#define OpcUaId_UnregisterNodesResponse_Encoding_DefaultXml 568
#define OpcUaId_QueryFirstRequest_Encoding_DefaultXml 614
#define OpcUaId_QueryFirstResponse_Encoding_DefaultXml 617
#define OpcUaId_QueryNextRequest_Encoding_DefaultXml 620
#define OpcUaId_QueryNextResponse_Encoding_DefaultXml 623
#define OpcUaId_ReadRequest_Encoding_DefaultXml 630
#define OpcUaId_ReadResponse_Encoding_DefaultXml 633
#define OpcUaId_HistoryReadRequest_Encoding_DefaultXml 663
#define OpcUaId_HistoryReadResponse_Encoding_DefaultXml 666
#define OpcUaId_WriteRequest_Encoding_DefaultXml 672
#define OpcUaId_WriteResponse_Encoding_DefaultXml 675
#define OpcUaId_HistoryUpdateRequest_Encoding_DefaultXml 699
#define OpcUaId_HistoryUpdateResponse_Encoding_DefaultXml 702
#define OpcUaId_CallRequest_Encoding_DefaultXml 711
#define OpcUaId_CallResponse_Encoding_DefaultXml 714
#define OpcUaId_CreateMonitoredItemsRequest_Encoding_DefaultXml 750
#define OpcUaId_CreateMonitoredItemsResponse_Encoding_DefaultXml 753
#define OpcUaId_ModifyMonitoredItemsRequest_Encoding_DefaultXml 762
#define OpcUaId_ModifyMonitoredItemsResponse_Encoding_DefaultXml 765
#define OpcUaId_SetMonitoringModeRequest_Encoding_DefaultXml 768
#define OpcUaId_SetMonitoringModeResponse_Encoding_DefaultXml 771
#define OpcUaId_SetTriggeringRequest_Encoding_DefaultXml 774
#define OpcUaId_SetTriggeringResponse_Encoding_DefaultXml 777
#define OpcUaId_DeleteMonitoredItemsRequest_Encoding_DefaultXml 780
#define OpcUaId_DeleteMonitoredItemsResponse_Encoding_DefaultXml 783
#define OpcUaId_CreateSubscriptionRequest_Encoding_DefaultXml 786
#define OpcUaId_CreateSubscriptionResponse_Encoding_DefaultXml 789
#define OpcUaId_ModifySubscriptionRequest_Encoding_DefaultXml 792
#define OpcUaId_ModifySubscriptionResponse_Encoding_DefaultXml 795
#define OpcUaId_SetPublishingModeRequest_Encoding_DefaultXml 798
#define OpcUaId_SetPublishingModeResponse_Encoding_DefaultXml 801
#define OpcUaId_PublishRequest_Encoding_DefaultXml 825
#define OpcUaId_PublishResponse_Encoding_DefaultXml 828
#define OpcUaId_RepublishRequest_Encoding_DefaultXml 831
#define OpcUaId_RepublishResponse_Encoding_DefaultXml 834
#define OpcUaId_TransferSubscriptionsRequest_Encoding_DefaultXml 840
#define OpcUaId_TransferSubscriptionsResponse_Encoding_DefaultXml 843
#define OpcUaId_DeleteSubscriptionsRequest_Encoding_DefaultXml 846
#define OpcUaId_DeleteSubscriptionsResponse_Encoding_DefaultXml 849
#define OpcUaId_TestStackRequest_Encoding_DefaultXml 409
#define OpcUaId_TestStackResponse_Encoding_DefaultXml 412
#define OpcUaId_TestStackExRequest_Encoding_DefaultXml 415
#define OpcUaId_TestStackExResponse_Encoding_DefaultXml 418

