/* packet-lte-rrc-template.h
 * Copyright 2009, Anders Broman <anders.broman@ericsson.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef PACKET_LTE_RRC_H
#define PACKET_LTE_RRC_H

extern value_string_ext lte_rrc_messageIdentifier_vals_ext;

#include "packet-lte-rrc-exp.h"

#endif  /* PACKET_LTE_RRC_H */
