/* pinfo_stats_tree.h
 * Stats tree for ethernet frames
 *
 *  (c) 2005, Luis E. G. Ontanon <luis@ontanon.org>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

extern void register_pinfo_stat_trees(void);
