/* packet-wlancertextn.h
 * Routines for Wireless Certificate Extensions (RFC3770) packet dissection
 *  Ronnie Sahlberg 2005
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef PACKET_WLANCERTEXTN_H
#define PACKET_WLANCERTEXTN_H

/*#include "packet-wlancertextn-exp.h"*/

#endif  /* PACKET_WLANCERTEXTN_H */

